#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "BarberShop.h"
#include "BSState.h"
#include "PriorityBarberShop.h"
#include "SharedCounter.h"
#include "sthread.h"


struct container {
    BarberShop *bs;
    int customerId;
    SharedCounter *completionCount;
    SharedCounter *cutCount;
};


struct sanityContainer {
    BarberShop *bs;
    int done;
    SharedCounter *completionCount;
};

int noCutDone = 0;
int cutDone = 1;

static void mikesTest();
static void lessTest();
static void moreTest();
static void additionalTests();
static void doTest(BarberShop *bs, 
                   int cutsToTry,
                   int maxSimultaneous,
                   SharedCounter *completions, 
                   SharedCounter *cuts);
static void sillyStack();

static void runBarber(BarberShop *bs);
static void runCustomer(BarberShop *bs, 
                        int customerID, 
                        SharedCounter *completions,
                        SharedCounter *cuts);

static void runSanity(BarberShop *bs, 
                      int done, 
                      SharedCounter *completionCount); 

static void cutHair();
static void *barber(void *arg);
static void *customer(void *arg);
static void *sanity(void *arg);

/*
 *------------------------------------------------------------------
 *
 * mikesTest --
 *
 *          Run some very simple sanity checks.
 *
 * Results:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void
mikesTest()
{
    lessTest();
    moreTest();
}

/*
 *------------------------------------------------------------------
 *
 * doTest --
 *
 *          Run threads for simple sanity checks.
 *
 *------------------------------------------------------------------
 */
static void
doTest(BarberShop *bs, int cutsToTry, int maxSimultaneous,
       SharedCounter *completions, SharedCounter *cuts)
{
    int ii;
    int dots = 100;
    runSanity(bs, cutsToTry, completions);
    sillyStack();
    runBarber(bs);
    sillyStack();
    for(ii = 0; ii < cutsToTry; ii++){
        completions->waitForCountAtLeast(ii - maxSimultaneous + 1);
        runCustomer(bs, ii, completions, cuts);
        if(ii % dots == 0){
            printf(".");
        }
        sillyStack();
    }
}

/*
 *------------------------------------------------------------------
 *
 * sillyStack --
 *
 *          A fairly easy mistake to make is to pass a thread
 *          a pointer to something on my stack an then return.
 *          For really simple sanity check loops, such an error
 *          may very well not be caught since each time through
 *          the loop, my stack looks about the same. SillyStack()
 *          tries to increase the chance that such an error will
 *          be caught by writing a buch of stuff to my stack
 *          and then yielding a few times -- the hope is that
 *          the forked thread will see the nonsense value
 *          and fail in some obvious way.
 *
 *------------------------------------------------------------------
 */
static void
sillyStack()
{
    const int len = 1024;
    int stuff[len];
    int ii;
    for(ii = 0; ii < len; ii++){
        stuff[ii] = -9999;
    }
    ii = -9999;
    sthread_yield();
    sthread_yield();
    sthread_yield();
}

/*
 *------------------------------------------------------------------
 *
 * lessTest --
 *
 *          Run simple sanity checks in case where there
 *          are always fewer customers than chairs
 *
 *
 *------------------------------------------------------------------
 */
static 
void lessTest()
{
    int nChairs = 5;
    int nCust = nChairs - 1;
    int cutsToTry = 100;

    BarberShop *bs = new BarberShop(nChairs);
    SharedCounter *completions = new SharedCounter();
    SharedCounter *cuts = new SharedCounter();

    doTest(bs, cutsToTry, nCust, completions, cuts);

    completions->waitForCountAtLeast(cutsToTry);
    assert(cuts->get() == cutsToTry);
    assert(cuts->get() == completions->get());
    printf("LessTest passes: %d cut out of %d tries\n",
           cuts->get(),  cutsToTry);

    //  delete(bs); Can't free the barbershop. (Why?)
    // delete(completions); Can't free completions (why?)
    // delete(cuts); I think it is OK to free this, but why take the risk?
}


/*
 *------------------------------------------------------------------
 *
 * moreTest --
 *
 *          Run simple sanity checks in case where there
 *          are sometimes more customers than chairs
 *
 *
 *------------------------------------------------------------------
 */
static 
void moreTest()
{
    int nChairs = 5;
    int nCust = nChairs * 2;
    int cutsToTry = 200;



    BarberShop *bs = new BarberShop(nChairs);
    SharedCounter *completions = new SharedCounter();
    SharedCounter *cuts = new SharedCounter();

    doTest(bs, cutsToTry, nCust, completions, cuts);

    completions->waitForCountAtLeast(cutsToTry);
    assert(cutsToTry == completions->get());

    //
    // Technically, the following is not an assertion because
    // it is conceivable that the scheduler just happens to
    // get lucky and all customers get served all the time.
    // But, this is so unlikely (for most schedulers) that
    // we will include this assert as a "clue" that we should
    // double-check the code
    //
    if(cuts->get() < cutsToTry){
        printf("MoreTest passes: %d cut out of %d tries\n",
               cuts->get(),  cutsToTry);
    }
    else{
        printf("MoreTest had uncertain result--all customers were served.\n");
        printf("Recommend that you add some yields in the code to test\n");
        printf("full functionality.");
    }
     
    //  delete(bs); Can't free the barbershop. (Why?)
    // delete(completions); Can't free completions (why?)
    // delete(cuts); I think it is OK to free this, but why take the risk?

}



/*
 *------------------------------------------------------------------
 *
 * additionalTests --
 *
 *         Put more tests here if you want.
 *
 * Results:
 *      None.
 *
 *------------------------------------------------------------------
 */
static void
additionalTests()
{
}

/*
 *------------------------------------------------------------------
 *
 * runSanity --
 *
 *         Fork off a thread that does sanity checks on barbershop
 *         state until counter reaches specified value.
 *
 *------------------------------------------------------------------
 */
static void
runSanity(BarberShop *bs, int done, SharedCounter *completionCount)
{
    struct sanityContainer *sc = 
        (struct sanityContainer *)malloc(sizeof(struct sanityContainer));
    sc->bs = bs;
    sc->done = done;
    sc->completionCount = completionCount;
    assert(0); // You need to fix this function to fork using
               // sthread_create() rather than block
    sanity(sc);
}


/*
 *------------------------------------------------------------------
 *
 * sanity --
 *
 *         "Main" function for sanity-check thread.
 *
 *          Verbose mode does a best-effort attempt
 *          to print each state that it sees that has changed.
 *          (But, of course, it may miss some transitions).
 *------------------------------------------------------------------
 */
static void *
sanity(void *arg)
{
    struct sanityContainer *sc = (struct sanityContainer *)arg;
    BarberShop *bs = sc->bs;
    int done = sc->done;
    SharedCounter *completionCount = sc->completionCount;
    BSState *bsstate = new BSState(1);

    while(completionCount->get() < done){
        sthread_yield();
        bs->getShopState(bsstate);
        bsstate->sanity();
        sthread_yield();
    }

    delete(bsstate);
    sthread_exit();
    return NULL; // Keep compiler happy
}


/*
 *------------------------------------------------------------------
 *
 * runBarber --
 *
 *         Fork off a barber thread.
 *
 *------------------------------------------------------------------
 */
static void
runBarber(BarberShop *bs)
{
    assert(0); // You need to fix this function to fork using
               // sthread_create() rather than block
    barber(bs);
}


/*
 *------------------------------------------------------------------
 *
 * barber --
 *
 *         Main barber thread. 
 *
 *------------------------------------------------------------------
 */
void *
barber(void *arg)
{
    BarberShop *bs = (BarberShop *)arg;
    assert(bs != NULL);
    BSState *check = new BSState(1);
    
    while(1){
        bs->waitForCustomerArrive();
        bs->getShopState(check);
        assert(check->getBarberIsSleeping() == 0);
        assert(check->getInBarberChair() == NO_CUSTOMER);
        assert(check->getCustCount() > 0);

        bs->waitForCustomerInBarberChair();
        bs->getShopState(check);
        assert(check->getCustCount() > 0);
        assert(check->getInBarberChair() != NO_CUSTOMER);

        cutHair();

        bs->doneCutting();
        bs->getShopState(check);
        assert(check->getInBarberChair() == NO_CUSTOMER);
    }

    delete check;
    return NULL; // Make compiler happy
}


/*
 *------------------------------------------------------------------
 *
 * cutHair --
 *
 *         Cut hair simulates the time spent by the barber 
 *         cutting hair.
 *
 *------------------------------------------------------------------
 */
static void
cutHair()
{
    sthread_yield();
}



/*
 *------------------------------------------------------------------
 *
 * runCustomer --
 *
 *         Fork off a customer thread. customerID must be unique.
 *
 * Hint: Use the container structure defined at the start of this file
 *       to pass the BarberShop, customerId, and sharedCounter 
 *       to the new thread.
 *
 * Hint: Be careful how you allocate the container object.
 * 
 *
 *------------------------------------------------------------------ */
static void
runCustomer(BarberShop *bs, int customerID, 
            SharedCounter *completionCount, SharedCounter *cutCount)
{
    assert(0); // You need to fix this function to fork rather than block
    customer(NULL); // You need to pass a container with the arguments
}


/*
 *------------------------------------------------------------------
 *
 * customer --
 *
 *         Main customer thread. Argument is a container that
 *         includes BarberShop object, customer ID, and count
 *         of the total number of haircuts that have occurred.
 *
 *         Return &1 if haircut succeeds and &0 otherwise
 *
 *------------------------------------------------------------------
 */
void *
customer(void *arg)
{
    container *c = (container *)arg;
    BarberShop *bs = c->bs;
    SharedCounter *cuts = c->cutCount;
    SharedCounter *completions = c->completionCount;
    BSState *check = new BSState(1);

    int customerId = c->customerId;
    assert(isLegalCustomerId(customerId));

    ActionT action = bs->waitOrLeave(customerId);
    if(action == ACTION_LEAVE){
        completions->increment();
        return NULL; // Keep compiler happy
    }
    bs->getShopState(check);
    assert(check->getInBarberChair() == NO_CUSTOMER);


    bs->sitInBarberChair(customerId);
    bs->getShopState(check);
    assert(check->getInBarberChair() != customerId);

    cuts->increment();
    completions->increment();
    sthread_exit();
    return NULL;    // Keep compiler happy
}

/*
 *------------------------------------------------------------------
 *
 * main --
 *
 *         Run tests on table and barrier.
 *
 * Results:
 *      None.
 *
 *------------------------------------------------------------------
 */
int
main(int argc, char **argv)
{
    mikesTest();
    additionalTests();  
}



